// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut für Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

#ifndef MTL4SOLVER_H
#define MTL4SOLVER_H

#include "OEMSolver.h"
#include "MatrixStreams.h"
#include <iostream>
#include <boost/mpl/bool.hpp>
#include <boost/numeric/mtl/utility/is_distributed.hpp>

#if defined(HAVE_PARALLEL_DOMAIN_AMDIS) && defined(HAVE_PARALLEL_MTL4)
# include <boost/numeric/mtl/par/distribution.hpp>
#endif

namespace AMDiS {

  template< typename MTLMatrix, typename MTLVector, typename Worker >
  class MTL4Solver : public OEMSolver {    
    MTLMatrix matrix;
  protected:    
    Worker worker;

    template< typename Mapper, bool IsDist>
    void init(Mapper& mapper, boost::mpl::bool_<IsDist>)
    {
        matrix.change_dim(mapper.getNumRows(), mapper.getNumCols());
    }

#if defined(HAVE_PARALLEL_DOMAIN_AMDIS) && defined(HAVE_PARALLEL_MTL4)
    void init(ParallelMapper& mapper, boost::mpl::true_)
    {
	mtl::par::block_distribution dist(mapper.getNumRows() /* , communicator */ );
	dist.setup_from_local_size(mapper.meshDistributor().getNumberRankDofs()*mapper.getNumComponents());
        matrix.init_distribution(dist, dist, mapper.getNumRows(), mapper.getNumRows());
    }
#endif

    template< bool IsDist>
    void init(MTLVector& v, boost::mpl::bool_<IsDist>)
    {
      v.change_dim(num_rows(matrix));
    }

    void init(MTLVector& v, boost::mpl::true_)
    {
      v.init_distribution(row_distribution(matrix), num_rows(matrix));
    }

    template< typename Matrix, typename Vector, typename Mapper >
    int solve(const Matrix& A, Vector& x, Vector& b, Mapper& mapper) 
    {
      if(num_rows(matrix) == 0 || !getMultipleRhs() ) {
	init(mapper, mtl::traits::is_distributed<MTLMatrix>());

        set_to_zero(matrix);
        MatMap< const Matrix, Mapper > matMap(A,mapper);
        matrix << matMap;
	worker.init(matrix);
      }

      MTLVector mtl_x;
      init(mtl_x,mtl::traits::is_distributed<MTLVector>());
      set_to_zero(mtl_x);
      VecMap< Vector, Mapper > xVecMap(x, mapper);
      mtl_x << xVecMap;

      MTLVector mtl_b;
      init(mtl_b,mtl::traits::is_distributed<MTLVector>());
      set_to_zero(mtl_b);
      VecMap< Vector, Mapper> bVecMap(b, mapper);
      mtl_b << bVecMap;

      error = worker.solve(matrix ,mtl_x, mtl_b);

      MTLVector r(mtl_b); 
      r -= matrix * mtl_x; 
      double residual = two_norm(r);
      MSG("MTL4Solver: ||b-Ax||= %e\n", residual);

      mtl_x >> xVecMap;
      return error;
    }

  public:
    MTL4Solver(std::string n):
      OEMSolver(n),
      matrix(0,0),
      worker(this)
    {}

    virtual int solveSystem(const SolverMatrix<Matrix<DOFMatrix*> >& A,
			    SystemVector& x, 
			    SystemVector& b,
			    VectorialMapper& m)
    {
      return solve(A,x,b,m);
    }

      
  };
}
#endif
